# Sherlock version 1.0.0
# SherlockLexer.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Sherlock.
# Sherlock is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Sherlock is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Sherlock.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Token
# 3) Lexer

# 1) Imported modules
# -------------------

import string
import sys
import re

# 2) Token
# --------

class SherlockToken:
  LBRACK = "LBRACK"
  RBRACK = "RBRACK"
  LBRACE = "LBRACE"
  RBRACE = "RBRACE"
  COMMA = "COMMA"
  SEMICOLON = "SEMICOLON"
  EQ = "EQ"
  DF = "DF"
  LT = "LT"
  GT = "GT"
  LEQ = "LEQ"
  GEQ = "GEQ"
  PLUS = "PLUS"
  MINUS = "MINUS"
  STAR = "STAR"
  INTEGER = "INTEGER"
  IDENTIFIER = "IDENTIFIER"
  KEYWORD_ALLDIFFERENT = "KEYWORD_ALLDIFFERENT"
  KEYWORD_CONSTRAINT = "KEYWORD_CONSTRAINT"
  KEYWORD_DOMAIN = "KEYWORD_DOMAIN"
  KEYWORD_END = "KEYWORD_END"
  KEYWORD_MAXIMIZE = "KEYWORD_MAXIMIZE"
  KEYWORD_MINIMIZE = "KEYWORD_MINIMIZE"
  KEYWORD_OR = "KEYWORD_OR"
  KEYWORD_PROBLEM = "KEYWORD_PROBLEM"
  COMMENT = "COMMENT"
  UNKNOWN = "UNKNOWN"
  NON_TERMINATED_COMMENT = "NON_TERMINATED_COMMENT"

  def __init__(self, type, string, fileName, lineNumber, columnNumber):
    self.type = type
    self.string = string
    self.fileName = fileName
    self.lineNumber = lineNumber
    self.columnNumber = columnNumber

  @staticmethod
  def GetNameOfType(type):
    return type

# 3) Lexer
# --------

class SherlockLexer:
  _Keywords = {
    "alldifferent": SherlockToken.KEYWORD_ALLDIFFERENT,
    "constraint": SherlockToken.KEYWORD_CONSTRAINT,
    "domain": SherlockToken.KEYWORD_DOMAIN,
    "end": SherlockToken.KEYWORD_END,
    "maximize": SherlockToken.KEYWORD_MAXIMIZE,
    "minimize": SherlockToken.KEYWORD_MINIMIZE,
    "or": SherlockToken.KEYWORD_OR,
    "problem": SherlockToken.KEYWORD_PROBLEM
    }
  END_OF_FILE = 0
  INITIAL = 1
  IDENTIFIER = 10
  INTEGER = 20
  SINGLE_LINE_COMMENT = 40
  MULTILINE_COMMENT_1 = 50
  MULTILINE_COMMENT_2 = 51

  def __init__(self):
    self.keepComments = False
    self.fileName = None
    self.file = None
    self.lineNumber = 0
    self.columnNumber = 0
    self.line = None
    self.endOfFile = 0
    self.tokenLineNumber = 0
    self.tokenColumnNumber = 0
    self.tokenString = None
    self.readChars = []
    self.tokens = []
    self.state = -1

  def Initialize(self, fileName, file):
    self.fileName = fileName
    self.file = file
    self.tokens = []
    self.line = None
    self.lineNumber = 0
    self.columnNumber = 0
    self.endOfFile = 0
    self.tokenLineNumber = 0
    self.tokenColumnNumber = 0
    self.tokenString = None
    self.readChars = []
    self.GetNewLine()
    self.state = SherlockLexer.INITIAL
    return 0

  def Finalize(self):
    pass
     # self.file.close()

  def GetToken(self, tokenIndex):
    if tokenIndex<len(self.tokens):
      return self.tokens[tokenIndex]
    while self.state!=SherlockLexer.END_OF_FILE and tokenIndex>=len(self.tokens):
      if self.state==SherlockLexer.INITIAL:
        c = self.GetChar()
        self.tokenLineNumber = self.lineNumber
        self.tokenColumnNumber = self.columnNumber-1
        if c==None:
          self.state = SherlockLexer.END_OF_FILE
        elif c.isspace():
          continue
        elif c=="[":
          self.tokenString = c
          self.NewToken(SherlockToken.LBRACK)
        elif c=="]":
          self.tokenString = c
          self.NewToken(SherlockToken.RBRACK)
        elif c=="{":
          self.tokenString = c
          self.NewToken(SherlockToken.LBRACE)
        elif c=="}":
          self.tokenString = c
          self.NewToken(SherlockToken.RBRACE)
        elif c==",":
          self.tokenString = c
          self.NewToken(SherlockToken.COMMA)
        elif c==";":
          self.tokenString = c
          self.NewToken(SherlockToken.SEMICOLON)
        elif c=="+":
          self.tokenString = c
          self.NewToken(SherlockToken.PLUS)
        elif c=="-":
          self.tokenString = c
          self.NewToken(SherlockToken.MINUS)
        elif c=="*":
          self.tokenString = c
          self.NewToken(SherlockToken.STAR)
        elif c=="=":
          c = self.GetChar()
          if c=="=":
            self.tokenString = "=="
            self.NewToken(SherlockToken.EQ)
          else:
            self.tokenString = "="
            self.PutCharBack(c)
            self.NewToken(SherlockToken.UNKNOWN)
        elif c=="!":
          c = self.GetChar()
          if c=="=":
            self.tokenString = "!="
            self.NewToken(SherlockToken.DF)
          else:
            self.tokenString = "!"
            self.PutCharBack(c)
            self.NewToken(SherlockToken.UNKNOWN)
        elif c=="<":
          c = self.GetChar()
          if c=="=":
            self.tokenString = "<="
            self.NewToken(SherlockToken.LEQ)
          else:
            self.tokenString = "<"
            self.PutCharBack(c)
            self.NewToken(SherlockToken.LT)
        elif c==">":
          c = self.GetChar()
          if c=="=":
            self.tokenString = ">="
            self.NewToken(SherlockToken.GEQ)
          else:
            self.tokenString = ">"
            self.PutCharBack(c)
            self.NewToken(SherlockToken.GT)
        elif c=="/":
          c = self.GetChar()
          if c=="/":
            self.tokenString = "//"
            self.state = SherlockLexer.SINGLE_LINE_COMMENT
          elif c=="*":
            self.tokenString = "/*"
            self.state = SherlockLexer.MULTILINE_COMMENT_1
          else:
            self.tokenString = "/"
            self.PutCharBack(c)
            self.NewToken(SherlockToken.UNKNOWN)
        elif c.isalpha() or c=="_":
          self.tokenString = c
          self.state = SherlockLexer.IDENTIFIER
        elif c.isdigit():
          self.tokenString = c
          self.state = SherlockLexer.INTEGER
        else:
          self.tokenString = c
          self.NewToken(SherlockToken.UNKNOWN)

      # -----------------------------------------------------
      # Identifiers
      # -----------------------------------------------------
      elif self.state==SherlockLexer.IDENTIFIER:
        c = self.GetChar()
        if c.isdigit() or c.isalpha() or c=="_":
          self.tokenString += c
        else:
          self.PutCharBack(c)
          type = self.GetIdentifierType(self.tokenString)
          self.NewToken(type)
          self.state = SherlockLexer.INITIAL

      # -----------------------------------------------------
      # Integers
      # -----------------------------------------------------
      elif self.state==SherlockLexer.INTEGER:
        c = self.GetChar()
        if c==None:
          self.NewToken(SherlockToken.INTEGER)
          self.state = SherlockLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        else:
          self.PutCharBack(c)
          self.NewToken(SherlockToken.INTEGER)
          self.state = SherlockLexer.INITIAL

      # -----------------------------------------------------
      # Single line comments
      # -----------------------------------------------------
      elif self.state==SherlockLexer.SINGLE_LINE_COMMENT:
        c = self.GetChar()
        if c==None:
          if self.keepComments:
            self.NewToken(SherlockToken.COMMENT)
          self.state = 0
        elif c=="\n":
          if self.keepComments:
            self.NewToken(SherlockToken.COMMENT)
          self.state = 1
        else:
          self.tokenString += c

      # -----------------------------------------------------
      # Multiline comments
      # -----------------------------------------------------
      elif self.state==SherlockLexer.MULTILINE_COMMENT_1:
        c = self.GetChar()
        if c==None:
          self.NewToken(SherlockToken.NON_TERMINATED_COMMENT)
          self.state = SherlockLexer.END_OF_FILE
        elif c=="*":
          self.tokenString += c
          self.state = SherlockLexer.MULTILINE_COMMENT_2
        else:
          self.tokenString += c
      elif self.state==SherlockLexer.MULTILINE_COMMENT_2:
        c = self.GetChar()
        if c==None:
          self.NewToken(SherlockToken.NON_TERMINATED_COMMENT)
          self.state = SherlockLexer.END_OF_FILE
        elif c=="*":
          self.tokenString += c
          c = self.GetChar()
        elif c=="/":
          self.tokenString += c
          if self.keepComments:
            self.NewToken(SherlockToken.COMMENT)
          self.state = SherlockLexer.INITIAL
        else:
          self.tokenString += c
          self.state = SherlockLexer.MULTILINE_COMMENT_1

    if tokenIndex>=len(self.tokens):
      return None
    return self.tokens[tokenIndex]

  def GetIdentifierType(self, identifier):
    return SherlockLexer._Keywords.get(identifier, SherlockToken.IDENTIFIER)

  def NewToken(self, type):
    token = SherlockToken(type, self.tokenString, self.fileName, self.tokenLineNumber, self.tokenColumnNumber)
    self.tokens.append(token)

  def GetChar(self):
    if len(self.readChars)!=0:
      return self.readChars.pop()
    if self.columnNumber==len(self.line):
      self.GetNewLine()
      if self.endOfFile:
        return None
      return self.GetChar()
    c = self.line[self.columnNumber]
    self.columnNumber += 1
    return c

  def PutCharBack(self, c):
    self.readChars.append(c)

  def GetNewLine(self):
    self.line = self.file.readline()
    if self.line=="":
      self.endOfFile = 1
      return
    self.lineNumber += 1
    self.columnNumber = 0
    self.line = self.line.rstrip()
    self.line += "\n"

  def ScanFile(self, fileName):
    error = self.Initialize(fileName)
    if error!=0:
      return
    tokenIndex = 0
    while True:
      token = self.GetToken(tokenIndex)
      if token==None:
        break
      tokenIndex += 1
    self.Finalize()

  def PrintTokens(self, outputFile):
    for token in self.tokens:
      outputFile.write(str(token.lineNumber) + "\t")
      outputFile.write(str(token.columnNumber) + "\t")
      outputFile.write(SherlockToken.GetNameOfType(token.type) + "\t")
      outputFile.write(token.string + "\n")
