# Sherlock version 1.0.0
# SherlockModels.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Sherlock, a primitive solver for numerical contraint satisfaction problems.
# Sherlock is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Sherlock is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Sherlock.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Domains
# 3) Variables
# 4) Constraints
#   4.1) Mother class
#   4.2) All different
#   4.3) Monomials
# 5) Problems
# 6) Model

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

# 2) Domains
# ----------

class SherlockDomain:
  SET = 1
  RANGE = 2
  def __init__(self):
    self.type = None
    self._name = None
    self._minimumValue = None
    self._maximumValue = None
    self._values = []

  @staticmethod
  def NewRange(minimumValue, maximumValue):
    domain = SherlockDomain()
    domain.type = SherlockDomain.RANGE
    domain._minimumValue = minimumValue
    domain._maximumValue = maximumValue
    domain._values = list(range(minimumValue, maximumValue+1))
    return(domain)

  @staticmethod
  def NewSet(values):
    domain = SherlockDomain()
    domain.type = SherlockDomain.SET
    for value in values:
      domain._values.append(int(value))
    domain._values.sort()
    domain._minimumValue = domain._values[0]
    domain._maximumValue = domain._values[-1]
    return(domain)

  def GetName(self):
    return self._name
  def SetName(self, name):
    self._name = name

  def GetMinimumValue(self):
    return self._minimumValue
  def GetMaximumValue(self):
    return self._maximumValue
  def GetSize(self):
    return len(self._values)
  def IsRange(self):
    return self.type==SherlockDomain.RANGE
  def IsSet(self):
    return self.type==SherlockDomain.SET

  def DoesContainValue(self, value):
    return value in self._values
  def GetValues(self):
    return self._values
  def GetFirstValue(self):
    if self.GetSize()==0:
      return None
    return self._values[0]
  def RemoveValue(self, value):
    self._values.remove(value)
    if self.GetSize()==0:
      self._minimumValue = 1
      self._maximumValue = 0
    else:
      if value==self._minimumValue:
        self._minimumValue = self._values[0]
      elif value==self._maximumValue:
        self._maximumValue = self._values[-1]

  def UpdateMinimumValue(self, lowerBound):
    if self._minimumValue < lowerBound:
      self._minimumValue = int(math.ceil(lowerBound))
      while len(self._values)>0 and self._values[0]<self._minimumValue:
        del self._values[0]
  def UpdateMaximumValue(self, upperBound):
    if self._maximumValue > upperBound:
      self._maximumValue = int(math.floor(upperBound))
      while len(self._values)>0 and self._values[-1]>self._maximumValue:
        del self._values[-1]
  def KeepOnlyValue(self, value):
    if not self.DoesContainValue(value):
      return
    self._minimumValue = value
    self._maximumValue = value
    self._values = [value]

  def Clone(self):
    clone = SherlockDomain()
    clone.type = self.type
    clone._minimumValue = self._minimumValue
    clone._maximumValue = self._maximumValue
    clone._values = self._values[:]
    return clone

# 3) Variables
# ------------

class SherlockVariable:
  def __init__(self, name, domain):
    self._name = name
    self._domain = domain
    self._occurrences = []
    self._stacked = False

  def GetName(self):
    return self._name

  def GetDomain(self):
    return self._domain

  def GetOccurrences(self):
    return self._occurrences
  def ResetOccurrenceList(self):
    self._occurrences = []
  def AddOccurrence(self, constraint):
    self._occurrences.append(constraint)

  def IsStacked(self):
    return self._stacked

  def SetStacked(self, value):
    self._stacked = value

# 4) Constraints
# --------------

# 4.1) Mother class
# -----------------

class SherlockConstraint:
  EQUAL = 1
  DIFFERENT = 2
  LESS = 3
  GREATER = 4
  LESS_OR_EQUAL = 5
  GREATER_OR_EQUAL = 6
  ALL_DIFFERENT = 7
  CLAUSE = 8
  def __init__(self, type):
    self._type = type

  def GetType(self):
    return self._type
  def IsInequality(self):
    return self._type==SherlockConstraint.EQUAL or self._type==SherlockConstraint.DIFFERENT or self._type==SherlockConstraint.LESS or self._type==SherlockConstraint.GREATER or self._type==SherlockConstraint.LESS_OR_EQUAL or self._type==SherlockConstraint.GREATER_OR_EQUAL
  def IsAlldDifferent(self):
    return self._type==SherlockConstraint.ALL_DIFFERENT
  def IsClause(self):
    return self._type==SherlockConstraint.CLAUSE

# 4.2) All different
# ------------------

class SherlockAllDifferent(SherlockConstraint):
  def __init__(self, variables):
    SherlockConstraint.__init__(self, SherlockConstraint.ALL_DIFFERENT)
    self._variables = variables

  def GetVariable(self, index):
    if index<0 or index>=len(self._variables):
      return None
    return self._variables[index]
  def GetVariables(self):
    return self._variables
  def DoesInvolveVariable(self, variable):
    return variable in self._variables

# 4.3) Monomials
# --------------

class SherlockMonomial:
  def __init__(self, coefficient, variable):
    self._coefficient = coefficient
    self._variable = variable

  def GetCoefficient(self):
    return self._coefficient
  def GetVariable(self):
    return self._variable

# 4.4) Inequalities
# -----------------

class SherlockInequality(SherlockConstraint):
  def __init__(self, type, polynomial, bound):
    SherlockConstraint.__init__(self, type)
    self._polynomial = polynomial
    self._bound = bound

  def GetPolynomial(self):
    return self._polynomial
  def GetBound(self):
    return self._bound

# 4.5) Clauses
# ------------

class SherlockLiteral:
  EQUAL = 1
  DIFFERENT = 2
  LESS = 3
  GREATER = 4
  LESS_OR_EQUAL = 5
  GREATER_OR_EQUAL = 6
  def __init__(self, type, variable, value):
    self._type = type
    self._variable = variable
    self._value = value

  def GetType(self):
    return self._type
  def GetVariable(self):
    return self._variable
  def GetValue(self):
    return self._value


class SherlockClause(SherlockConstraint):
  def __init__(self, literals):
    SherlockConstraint.__init__(self, SherlockConstraint.CLAUSE)
    self._literals = literals

  def GetLiterals(self):
    return self._literals

# 5) Problems
# -----------

class SherlockProblem:
  def __init__(self, name):
    self._name = name
    self._variables = dict()
    self._constraints = []
    self._objectivePolynomial = None
    self._isMinimizationProblem = True

  def GetName(self):
    return self._name

  def GetVariable(self, name):
    return self._variables.get(name, None)
  def AddVariable(self, name, domain):
    variable = SherlockVariable(name, domain)
    self._variables[name] = variable
    return variable
  def GetVariables(self):
    return self._variables.values()
  def GetSortedVariables(self):
    return [self.GetVariable(name) for name in sorted(self._variables)]
  def GetNumberOfVariables(self):
    return len(self._variables)

  def GetConstraints(self):
    return self._constraints
  def GetNumberOfConstraints(self):
    return len(self._constraints)
  def AddAllDifferent(self, variables):
    constraint = SherlockAllDifferent(variables)
    self._constraints.append(constraint)
    return constraint
  def AddClause(self, literals):
    constraint = SherlockClause(literals)
    self._constraints.append(constraint)
    return constraint
  def AddInequality(self, type, polynomial, bound):
    constraint = SherlockInequality(type, polynomial, bound)
    self._constraints.append(constraint)
    return constraint

  def SetObjectivePolynomial(self, polynomial, isMinimizationProblem):
    self._objectivePolynomial = polynomial
    self._isMinimizationProblem = isMinimizationProblem
  def GetObjectivePolynomial(self):
    return self._objectivePolynomial
  def IsMinimizationProblem(self):
    return self._isMinimizationProblem

  def BuildOccurrenceLists(self):
    for variable in self._variables.values():
      variable.ResetOccurrenceList()
    for constraint in self._constraints:
      if constraint.GetType()==SherlockConstraint.ALL_DIFFERENT:
        for variable in constraint.GetVariables():
          variable.AddOccurrence(constraint)
      elif constraint.GetType()==SherlockConstraint.CLAUSE:
        for literal in constraint.GetLiterals():
          variable = literal.GetVariable()
          variable.AddOccurrence(constraint)
      else:
        for monomial in constraint.GetPolynomial():
          variable = monomial.GetVariable()
          variable.AddOccurrence(constraint)

# 6) Model
# --------

class SherlockModel:
  def __init__(self):
    self._domains = dict()
    self._problems = dict()

  def GetDomain(self, name):
    return self._domains.get(name, None)
  def AddDomain(self, name, domain):
    self._domains[name] = domain
  def GetDomains(self):
    return self._domains.values()
  def GetSortedDomains(self):
    return [self.GetDomain(name) for name in sorted(self._domains)]
  def GetDomainNames(self):
    return self._domains.keys()
  def GetNumberOfDomains(self):
    return len(self._domains)

  def GetProblem(self, name):
    return self._problems.get(name, None)
  def AddProblem(self, name, problem):
    self._problems[name] = problem
  def GetProblems(self):
    return self._problems.values()
  def GetSortedProblems(self):
    return [self.GetProblem(name) for name in sorted(self._problems)]
  def GetProblemNames(self):
    return self._problems.keys()
  def GetNumberOfProblems(self):
    return len(self._problems)
