# Sherlock version 1.0.0
# SherlockPrinter.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Sherlock, a primitive solver for numerical contraint satisfaction problems.
# Sherlock is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Sherlock is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Sherlock.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Printer

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from SherlockModels import *
from SherlockSolver import SherlockAssignment

# 2) Printer
# ----------

class SherlockPrinter:
  def __init__(self):
    self._depth = 0
    self._indentation = "  "

  def PrintIndentation(self, output):
    for i in range(self._depth):
      output.write(self._indentation)

  def ExportModel(self, model, fileName):
    try:
      output = open(fileName, "w")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return
    self.PrintModel(model, output)
    output.close()

  def PrintModel(self, model, output):
    self._depth = 0
    if model.GetNumberOfDomains()!=0:
      for domain in model.GetSortedDomains():
        self.PrintDomainDeclaration(domain, output)
    if model.GetNumberOfProblems()!=0:
      for problem in model.GetSortedProblems():
        self.PrintProblem(problem, output)

  def PrintDomainDeclaration(self, domain, output):
    self.PrintIndentation(output)
    output.write("domain ")
    output.write("%s " % domain.GetName())
    self.PrintDomain(domain, output)
    output.write("\n")

  def PrintProblem(self, problem, output):
    self._depth = 0
    output.write("problem %s\n" % problem.GetName())
    self._depth +=1
    if problem.GetNumberOfVariables()!=0:
      for variable in problem.GetSortedVariables():
        self.PrintIndentation(output)
        domain = variable.GetDomain()
        if domain.GetName()!=None:
          output.write("%s" % domain.GetName())
        else:
          self.PrintDomain(domain, output)
        output.write(" %s;" % variable.GetName())
        output.write("\n")
    if problem.GetNumberOfConstraints()!=0:
      self.PrintIndentation(output)
      output.write("constraint\n")
      self._depth +=1
      for constraint in problem.GetConstraints():
        self.PrintIndentation(output)
        self.PrintConstraint(constraint, output)
        output.write("\n")
      self._depth -=1
    if problem.GetObjectivePolynomial()!=None:
      self.PrintIndentation(output)
      if problem.IsMinimizationProblem():
        output.write("minimize ")
      else:
        output.write("maximize ")
      self.PrintPolynomial(problem.GetObjectivePolynomial(), output)
      output.write(";\n")
    output.write("end\n")
    self._depth -=1
    output.flush()

  def PrintDomain(self, domain, output):
    if domain.IsRange():
      output.write("[%d, %d]" % (domain.GetMinimumValue(), domain.GetMaximumValue()))
    else:
      self.PrintDomainAsSet(domain, output)

  def PrintDomainAsSet(self, domain, output):
    output.write("{")
    first = True
    for value in domain.GetValues():
      if first:
        first = False
      else:
        output.write(", ")
      output.write(str(value))
    output.write("}")

  def PrintConstraint(self, constraint, output):
    if constraint.IsInequality():
      self.PrintInequality(constraint, output)
    elif constraint.IsAlldDifferent():
      self.PrintAllDifferent(constraint, output)
    elif constraint.IsClause():
      self.PrintClause(constraint, output)

  def PrintAllDifferent(self, constraint, output):
    output.write("alldifferent ")
    first = True
    for variable in constraint.GetVariables():
      if first:
        first = False
      else:
        output.write(", ")
      output.write("%s" % variable.GetName())
    output.write(";")

  def PrintClause(self, constraint, output):
    output.write("or ")
    first = True
    for literal in constraint.GetLiterals():
      if first:
        first = False
      else:
        output.write(", ")
      self.PrintLiteral(literal, output)
    output.write(";")

  def PrintLiteral(self, literal, output):
    output.write(literal.GetVariable().GetName())
    if literal.GetType()==SherlockLiteral.EQUAL:
      output.write(" == ")
    elif literal.GetType()==SherlockLiteral.DIFFERENT:
      output.write(" != ")
    elif literal.GetType()==SherlockLiteral.LESS:
      output.write(" < ")
    elif literal.GetType()==SherlockLiteral.GREATER:
      output.write(" > ")
    elif literal.GetType()==SherlockLiteral.LESS_OR_EQUAL:
      output.write(" <= ")
    elif literal.GetType()==SherlockLiteral.GREATER_OR_EQUAL:
      output.write(" >= ")
    output.write("%d" % literal.GetValue())

  def PrintInequality(self, constraint, output):
    self.PrintPolynomial(constraint.GetPolynomial(), output)
    if constraint.GetType()==SherlockConstraint.EQUAL:
      output.write(" == ")
    elif constraint.GetType()==SherlockConstraint.DIFFERENT:
      output.write(" != ")
    elif constraint.GetType()==SherlockConstraint.LESS:
      output.write(" < ")
    elif constraint.GetType()==SherlockConstraint.GREATER:
      output.write(" > ")
    elif constraint.GetType()==SherlockConstraint.LESS_OR_EQUAL:
      output.write(" <= ")
    elif constraint.GetType()==SherlockConstraint.GREATER_OR_EQUAL:
      output.write(" >= ")
    output.write("%d;" % constraint.GetBound())

  def PrintPolynomial(self, polynomial, output):
    first = True
    for monomial in polynomial:
      if monomial.GetCoefficient()<0:
        if first:
          first = False
          output.write("-")
        else:
          output.write(" - ")
        if monomial.GetCoefficient()!=1:
          output.write("%d*" % -monomial.GetCoefficient())
      else:
        if first:
          first = False
        else:
          output.write(" + ")
        if monomial.GetCoefficient()!=1:
          output.write("%d*" % monomial.GetCoefficient())
      output.write("%s" % monomial.GetVariable().GetName())

  def ExportAssignment(self, problem, assignment, fileName):
    try:
      output = open(fileName, "w")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return
    self.PrintAssignment(problem, assignment, output)
    output.close()

  def PrintAssignment(self, problem, assignment, output):
    output.write("problem %s\n" % problem.GetName())
    for variable in assignment.GetSortedVariables():
      output.write("\t%s " % variable.GetName())
      domain = assignment.GetDomain(variable)
      self.PrintDomainAsSet(domain, output)
      output.write("\n")
    output.write("end\n")
    output.flush()

  def PrintOccurrenceLists(self, problem, output):
    self.depth = 0
    for variable in problem.GetVariables():
      output.write("occurrences of %s\n" % variable.GetName())
      self._depth += 1
      for constraint in variable.GetOccurrences():
        self.PrintIndentation(output)
        self.PrintConstraint(constraint, output)
        output.write("\n")
      self._depth -= 1
